
cdef class typedarray:

	cdef void *ptr
	cdef size_t dims
	cdef size_t *shape
	cdef ssize_t *strides
	
	cdef DDType dtype
	cdef readonly owner
	
	@property
	def shape(self):
		''' number of element along each dimension '''
		cdef typedlist shape = typedlist.__new__()
		shape.ptr = self.shape
		shape.size = shape.allocated = self.dims*sizeof(size_t)
		shape.dtype = declared('N')
		shape.owner = self
		
	@property
	def strides(self):
		''' byte offset between elements along each dimension '''
		cdef typedlist shape = typedlist.__new__()
		shape.ptr = self.strides
		shape.size = shape.allocated = self.dims*sizeof(ssize_t)
		shape.dtype = declared('n')
		shape.owner = self
	
	@property
	def dtype(self):
		''' the python dtype object, or the ddtype if there is not dtype '''
		return self.dtype.key or self.dtype
		
	@property
	def ddtype(self):
		''' the declaration of the dtype, a DDType instance '''
		return self.dtype
	
	def __init__(self, buffer, shape, strides, dtype=None):
		assert len(shape) == len(strides)
		self.dims = len(shape)
		self.shape = <size_t*> PyMem_Malloc(2*self.dims*sizeof(size_t))
		self.strides = self.shape + self.dims
		self.dtype = declared(dtype)
		self.owner = buffer

		indev()
		
	def __del__(self):
		PyMem_Free(self.shape)
		self.shape = 0
		self.strides = 0
		
	cdef object _getitem(self, void* place):
		return self.dtype.c_unpack(self.dtype, place)
		
	cdef int _setitem(self, void* place, obj) except -1:
		return self.dtype.c_pack(self.dtype, place, obj)
		
	
	def __len__(self):
		if self.dims:
			return self.shape[0]
		else:
			return 0
	
	def __getitem__(self, *index):
		
		# try to extract a simple element
		# indices must all be int and there must be as many indices as there is dimensions
		if len(index) == self.dims:
			for i in range(len(index)):
				if PyNumber_Check(index[i]):
					if index[i] < 0 or self.shape[i] <= index[i]:
						raise IndexError('index out of range for dimension {}'.format(i))
					ptr += index[i] * self.strides[i]
				else:
					break
			else:
				return self._getitem(ptr)
		
		# else get a slice array
		cdef ssize_t i
		cdef Py_ssize_t start, stop, step
		cdef typedarray new = typedarray.__new__()
		
		new.dims = 0
		new.ptr = self.ptr
		new.shape = <size_t> PyMem_Malloc(2*self.dims*sizeof(size_t))
		new.strides = new.shape + self.dims
		new.dtype = self.dtype
		new.owner = self.owner
		
		cdef size_t indices = len(index)
		for i in range(self.dims):
			if i < len(index):
				if PyNumber_Check(index[i]):
					new.ptr += index[i] * self.strides[i]
				elif isinstance(index[i], slice):
					if PySlice_Unpack(index[i], &start, &stop, &step):
						raise IndexError('incorrect slice')
					PySlice_AdjustIndices(self.shape[i], &start, &stop, step)
					
					new.strides[dims] = self.strides[i] * step
					new.shape[dims] = stop - start
					new.ptr += start * self.strides[i]
					new.dims += 1
			else:
				new.strides[dims] = self.strides[i]
				new.shape[dims] = self.shape[i]
				new.dims += 1
				
		return new
		
			
		
	def __setitem__(self, index, value):
		indev()
		
	def __iter__(self):
		indev()
	
	def transpose(self, reindex=None):
		indev()
		
	def ravel(self, dims=None):
		indev()
		
	def reshape(self, shape):
		indev()
		
	def extract(self, indices):
		indev()
		
	def swap(self, other):
		indev()
	
	
	def __eq__(self, other):
		indev()
		
	def index(self, value):
		indev()
	
	
	def __copy__(self):
		indev()
		
	def __deepcopy__(self):
		indev()
	
	



import glm
from .dtypes import declare, DDTypeExtension
from struct import calcsize

native = 8	# for 64bit machines only

for prec, fmt in (
			('u8', 'B'), ('i8', 'b'), ('u16','H'), ('i16','h'), ('u64','Q'), ('i64','q'), 
			('u','I'), ('i','i'), 
			('f','f'), ('d','d')):
	for size in range(2,5):
		type = getattr(glm, prec+'vec'+str(size))
		pack = size*fmt
		remain = calcsize(pack) % native
		if remain:	pack += 'x' * (native - remain)
		declare(type, DDTypeExtension(type, pack, type))

for prec, fmt in (('u','I'), ('i','i'), ('f','f'), ('d','d')):
	for size in range(2,5):
		type = getattr(glm, prec+'mat'+str(size))
		declare(type, DDTypeExtension(type, size*fmt, type))
		
for prec in ('f', 'd'):
	type = getattr(glm, prec+'quat')
	declare(type, DDTypeExtension(type, 4*prec, type))



# cython: language_level=3, cdivision=True

'''
	The *dtype* is the type of the elements in a buffer. Thanks to the ddtype system, it is very easy to create new dtypes on top of pretty much everything.
	
	Definitions:
	
		:type:    a python type object (typically a class or a builtin type)
		:dtype:   data dtype, meaning the type of the elements in an array,  it can be a type, but more generally anything that define a data format.
		:ddtype:  declaration of data type, meaning a packet of things decribing how to pack/unpack that dtype from/to an array
		
	a ddtype always inherits from base class `DDType` which content is implemented at C level.
'''

cimport cython
from cpython cimport PyObject, PyTypeObject
from libc.string cimport memcpy

import struct, ctypes

cdef extern from "Python.h":
	object PyBytes_FromStringAndSize(const char *v, Py_ssize_t len)
	object PyByteArray_FromStringAndSize(const char *v, Py_ssize_t len)
	char *PyBytes_AsString(object)


	

cdef class DDType:
	''' base class for a declaration of data type (ddtype)
		DO NOT INSTANTIATE THIS CLASS FROM PYTHON, use on of its specialization instead 
		
		Attributes:
		
			dsize (int):     byte size of the dtype when packed
			layout (bytes):  layout of the packed data such as defined in module `struct`, or `None` if not defined
			key:             the python dtype itself if this DDType is declared, `None` if not declared
	'''
	
	def __repr__(self):
		if isinstance(self.key, type):
			return '<dtype {}>'.format(self.key.__name__)
		elif isinstance(self.key, str):
			return '<dtype {}>'.format(repr(self.key))
		else:
			return '<dtype at {}>'.format(id(self))
			
	def __reduce_ex__(self, protocol):
		''' allow serialization of the dtype with the array (particularly useful for anonymous dtypes) '''
		if self.key is None:
			raise TypeError('a dtype must be declared in order to be pickleable')
		return self._rebuild, (self.key, self.dsize, self.layout)
		
	@classmethod
	def _rebuild(cls, key, size_t dsize, bytes layout):
		candidate = declared(key)
		if dsize != candidate.dsize:
			raise ValueError('the pickled dtype {} has a different size here than in dump, unpickled {} expected {}'
						.format(repr(key), dsize, candidate.dsize))
		if candidate.layout and layout != candidate.layout:
			raise ValueError('the pickled dtype {} has a different memory layout here than in dump, unpickled {} expected {}'
						.format(repr(key), layout, candidate.layout))
		return candidate
	

def DDTypeClass(type):
	''' DDTypeClass(type)
	
		Create a dtype from a python class (can be a pure python class) 
		
		the given type must have the following attributes:
		
			- `frombytes` or `from_bytes` or `from_buffer`
			
				static method that initialize the type from bytes
				
			- `__bytes__` or `tobytes` or `to_bytes`
				
				method that converts to bytes, the returned byte must always be of the same size
			
			- `__packlayout__`     (optional)  string or bytes giving binary format returned by `__bytes__`, it must follow the specifications of module `struct`
			- `__packsize__`       (optional)  defines the byte size returned by `__bytes__`, optional if `__packlayout__` is provided
			
		Example:
		
			>>> class test_class:
			... 	__packlayout__ = 'ff'
			... 	_struct = struct.Struct(__packlayout__)
			... 	
			... 	def __init__(self, x, y):
			... 		self.x = x
			... 		self.y = y
			... 	
			... 	def __bytes__(self):
			... 		return self._struct.pack(self.x, self.y)
			... 	@classmethod
			... 	def frombytes(cls, b):
			... 		return cls(*cls._struct.unpack(b))
			... 		
			... 	def __repr__(self):
			... 		return '(x={}, y={})'.format(self.x, self.y)
			... 
			>>> a = typedlist(dtype=test_class)    # no declaration needed
	'''
	layout = getattr(type, '__packlayout__', None)
	dsize = getattr(type, '__packsize__', None)
	if not dsize:
		dsize = struct.calcsize(layout or b'')
	if not dsize:
		raise ValueError('dsize must not be null, __packlayout__ or __packsize__ must be correctly defined in the given type')
	
	pack = getattr(type, '__bytes__', None) or getattr(type, 'tobytes', None) or getattr(type, 'to_bytes', None)
	if not pack:
		raise TypeError("the given type must have a method '__bytes__', 'tobytes', or 'to_bytes'")
	
	unpack = getattr(type, 'frombytes', None) or getattr(type, 'from_bytes', None) or getattr(type, 'from_buffer', None)
	if not unpack:
		raise TypeError("the given type must have a method 'frombytes', 'from_bytes', or 'from_buffer'")
	
	return DDTypeFunctions(dsize, pack, unpack, layout)
	
def DDTypeStruct(definition):
	''' DDTypeStruct(struct)
	
		create a dtype from a Struct object from module `struct` 
	
		Example:
			
			>>> a = typedlist(dtype='fxBh')   # no declaration needed
	'''
	if not isinstance(definition, struct.Struct):
		raise TypeError('structure must be struct.Struct')
	return DDTypeFunctions(definition.size, lambda o: definition.pack(*o), definition.unpack, definition.format) 

cdef class DDTypeFunctions(DDType):
	''' DDTypeFunctions(dsize, pack, unpack, layout=None)
	
		create a dtype from pure python pack and unpack functions 
	
		Example:
		
			>>> enum_pack = {'apple':b'a', 'orange':b'o', 'cake':b'c'}
			>>> enum_unpack = {v:k   for k,v in enum_direct.items()}
			>>> enum_dtype = DDTypeFunctions(
			... 			dsize=1,                         # 1 byte storage
			... 			pack=enum_pack.__getitem__,      # this takes the python object and gives a bytes to dump
			... 			unpack=enum_unpack.__getitem__,  # this takes the bytes and return a python object
			... 			)
			... 
			>>> a = typedlist(dtype=enum_dtype)		# declaration is not necessary
	'''
	cdef public object pack
	cdef public object unpack
	
	def __init__(self, dsize, pack, unpack, layout=None):
		if not callable(pack) or not callable(unpack):
			raise TypeError('pack and unpack must be callables')
			
		if layout is not None:
			if isinstance(layout, str):
				layout = layout.encode()
			elif not isinstance(layout, bytes):
				layout = bytes(layout)
			
			fmtsize = struct.calcsize(layout)
			if dsize is not None and dsize != fmtsize:
				raise ValueError('dsize must match layout size')
		elif not dsize:
			raise ValueError('dsize must be provided or deduced from a given layout')
			
		self.dsize = dsize
		self.c_pack = <c_pack_t> self._func_pack
		self.c_unpack = <c_unpack_t> self._func_unpack
		self.layout = layout
		self.pack = pack
		self.unpack = unpack
		
	cdef int _func_pack(self, void* place, object obj) except -1:
		packed = self.pack(obj)
		if not isinstance(packed, bytes):
			raise TypeError('pack must provide a bytes object')
		if len(packed) < <ssize_t> self.dsize:
			raise ValueError('the dumped bytes length {} does not match dsize {}'.format(len(packed), self.dtype.dsize))
		
		memcpy(place, PyBytes_AsString(packed), self.dsize)
	
	cdef object _func_unpack(self, void* place):
		return self.unpack(PyBytes_FromStringAndSize(<char*>place, self.dsize))
		
		
	def __reduce_ex__(self, protocol):
		''' allow serialization of the dtype with the array (particularly useful for anonymous dtypes) '''
		return type(self), (self.dsize, self.pack, self.unpack, self.layout, self.constructor)
		
cdef class DDTypeCType(DDType):
	''' DDTypeCType(type)
	
		Create a dtype from a ctype 
	
		Example:
		
			>>> class test_structure(ctypes.Structure):
			... 	_fields_ = [
			... 		('x', ctypes.c_int),
			... 		('y', ctypes.c_float),
			... 		]
			... 	def __repr__(self):
			... 		return '(x={}, y={})'.format(self.x, self.y)
			... 
			>>> a = typedlist(dtype=test_structure)
	'''
	cdef public object type
	
	def __init__(self, type):
		if not isctype(type):
			raise TypeError('type must be a ctype')
		self.type = type
		self.dsize = ctypes.sizeof(type)
		self.c_pack = <c_pack_t> self._ctype_pack
		self.c_unpack = <c_unpack_t> self._ctype_unpack
		self.layout = None
		
	cdef int _ctype_pack(self, void* place, object obj) except -1:
		if not isinstance(obj, self.type):
			raise TypeError('cannot store an object of a different type')
		memcpy(place, <void*><size_t> ctypes.addressof(obj), self.dsize)
		
	cdef object _ctype_unpack(self, void* place):
		return self.type.from_buffer(PyByteArray_FromStringAndSize(<char*> place, self.dsize))
		
		
cdef class DDTypeExtension(DDType):
	''' DDTypeTypeExtension(type, layout=None, constructor=None)
	
		Create a dtype for a C extension type.
	
		This is the most efficient kind of dtype in term of access/assignation time.
		
		In order to put an extension object into an array, it satisfy the following conditions:
		
		- have fixed size known at the time of dtype creation (so any array element has the same)
		- contain only byte copiable data (so nothing particular is done when copying/destroying the objects)
		
		WARNING:  
		
			These conditions MUST be ensured by the user when declaring an extension type as a dtype, or it will result in memory corruption and crash of the program
		
		
		Example:
		
			>>> arrex.declare(vec3, DDTypeExtension(vec3, 'fff', vec3))
	'''
	cdef public type type
	cdef public object constructor
	
	def __init__(self, type ext, layout=None, constructor=None):
		cdef ssize_t fmtsize, packsize
	
		if not isinstance(ext, type):
			raise TypeError('dtype must be a type')
			
		packsize = (<PyTypeObject*> ext).tp_basicsize - sizeof(_head)
		if layout is not None:
			if isinstance(layout, str):
				layout = layout.encode()
			elif not isinstance(layout, bytes):
				layout = bytes(layout)
			
			fmtsize = struct.calcsize(layout)
			if packsize < fmtsize:
				raise ValueError('format describes a too big structure for the given dtype')
		else:
			fmtsize = 0
		
		self.dsize = fmtsize or packsize
		if not self.dsize:
			raise TypeError('dsize cannot be 0')
		
		self.c_pack = <c_pack_t> self._ext_pack
		self.c_unpack = <c_unpack_t> self._ext_unpack
		self.layout = layout
		self.type = ext
		self.constructor = constructor
		
	cdef void * _raw(self, obj):
		return (<char*><PyObject*> obj) + (<PyTypeObject*>self.type).tp_basicsize - self.dsize
	
	cdef int _ext_pack(self, void* place, object obj) except -1:
		if not isinstance(obj, self.type):
			if self.constructor is not None:
				obj = self.constructor(obj)
			else:
				raise TypeError('cannot implicitely convert {} into {}'.format(
										type(obj).__name__, 
										repr(self),
										))
		memcpy(place, self._raw(obj), self.dsize)
		
	cdef object _ext_unpack(self, void* place):
		new = (<PyTypeObject*>self.type).tp_new(self.type, _empty, None)
		memcpy(self._raw(new), place, self.dsize)
		return new
	
	def __reduce_ex__(self, protocol):
		''' that overload allows for automatic declaration of the dtype fron the pickled informations, as long as they are consistent with the given type size '''
		return self._rebuild, (self.type, self.dsize, self.layout, self.constructor)
		
	@classmethod
	def _rebuild(cls, type, size_t dsize, bytes layout, constructor):
		candidate = _declared.get(type) or declare(type, DDTypeExtension(type, layout, constructor))
		if dsize != candidate.dsize:
			raise ValueError('the pickled dtype {} has a different size here than in dump, unpickled {} expected {}'
						.format(repr(type), dsize, candidate.dsize))
		if candidate.layout and layout != candidate.layout:
			raise ValueError('the pickled dtype {} has a different memory layout here than in dump, unpickled {} expected {}'
						.format(repr(type), layout, candidate.layout))
		return candidate

		
		

# dictionnary of compatible packed types
cdef dict _declared = {}	# {python type: dtype}
	
cpdef DDType declare(key, DDType dtype=None):
	''' declare(dtype, ddtype)
	
		declare a new dtype 
	'''
	if not dtype:
		if isinstance(key, str):
			# create a struct dtype
			dtype = DDTypeStruct(struct.Struct(key))
		elif isctype(key):
			# create a ctype dtype
			dtype = DDTypeCType(key)
		elif isinstance(key, type):
			# create a dtype from a pure python class
			try:	dtype = DDTypeClass(key)
			except TypeError:	pass
		if not dtype:
			raise TypeError('dtype {} is not declared, and cannot be guessed'.format(key))
	if not dtype.key:
		dtype.key = key
	_declared[key] = dtype
	return dtype
	
cpdef DDType declared(key):
	''' declared(key)
		
		return the content of the declaration for the givne dtype 
	'''
	if isinstance(key, DDType):
		return key
	else:
		# try an automated declaration
		dtype = _declared.get(key) or declare(key)
		# raise an error when not declared
		#raise TypeError('dtype {} is not declared'.format(key))
		return <DDType> dtype

cdef isctype(obj):
	''' isctype(obj)
	
		return True if obj is a ctype type or one of its derivatives, and False otherwise 
	'''
	# there is currently no other way to check this than to try a ctype-only function and detect errors
	# this is an ungly way but better would require changes in ctypes ...
	if not isinstance(obj, type):	return False
	try:	ctypes.sizeof(obj)
	except TypeError:	return False
	return True


# create an empty object to easily get the PyObject head size
cdef class _head:
	''' implementation purpose only '''
	pass

# empty tuple, reused to fasten some calls
cdef tuple _empty = ()




# cython: language_level=3, cdivision=True

cimport cython
from cpython cimport PyObject, PyTypeObject, Py_buffer, Py_INCREF, Py_DECREF, Py_XDECREF, PyObject_Length
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from cpython.buffer cimport PyBUF_SIMPLE, PyBUF_ND, PyBUF_FORMAT, PyObject_CheckBuffer, PyObject_GetBuffer, PyBuffer_Release
from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AsString, PyBytes_AS_STRING
from cpython.slice cimport PySlice_Unpack, PySlice_AdjustIndices
from cpython.number cimport PyNumber_Check
from libc.string cimport memcpy, memmove, memcmp
from libc.stdlib cimport malloc, free

import struct
from pickle import PickleBuffer
import sys

cdef extern from "Python.h":
	Py_ssize_t PyObject_LengthHint(object o, Py_ssize_t default)


from .dtypes cimport *

# this is to avoid the issue around Py_buffer.obj = pyobject, which in cython would try to decref the initially NULL value
# so the value needs to be set using that function when set the first time
cdef extern from *:
	"""
	void assign_buffer_obj(Py_buffer* buf, PyObject* o) {
		Py_INCREF(o);
		buf->obj = o;
	}
	"""
	void assign_buffer_obj(Py_buffer* buf, object o)
	
cdef extern from *:
	"""
	void release_buffer_obj(Py_buffer* buf) {
		if (buf->obj != NULL)
			Py_DECREF(buf->obj);
	}
	"""
	void release_buffer_obj(Py_buffer* buf)


	
cdef class typedlist:
	''' list-like array that stores objects as packed data. 
		The objects added must necessarily be packed objects (builtin objects with no references).
		
		This is a dynamically sized, borrowing array, which mean the internal buffer of data is reallocated on insertion, but can be used to view and extract from any buffer.
		
		Methods added to the signature of list:
		
			`reserve(n)`            reallocate if necessary to make sure n elements can 
			                        be inserted without reallocation
			`capacity() -> int`     return the current number of elements that can be 
			                        contained without reallocation
			`shrink()`              shorten the allocated memory to fit the current content
			
			also the slices do not copy the content
		
		
		Use it as a list:
		
			>>> a = typedlist(dtype=vec3)
			>>> 
			>>> # build from an iterable
			>>> a = typedlist([], dtype=vec3)
			>>> 
			>>> # append some data
			>>> a.append(vec3(1,2,3))
			>>> 
			>>> # extend with an iterable
			>>> a.extend(vec3(i)  for i in range(5))
			>>> 
			>>> len(a)	# the current number of elements
			6
			>>> 
			>>> a.owner	# the current data buffer
			b'.........'
			
		Use it as a slice:
		
			>>> # no data is copied
			>>> myslice = a[:5]
			typedlist(....)
			
		Use it as a view on top of a random buffer
		
			>>> a = np.ones((6,3), dtype='f4')
			>>> myslice = typedlist(a, dtype=vec3)
			
		It does support the buffer protocol, so it can be converted in a great variety of well known arrays, even without any copy
		
			>>> np.array(typedlist([....]))
		
		
		Constructors:
			
			.. py:function:: typedlist()
			.. py:function:: typedlist(dtype, reserve=None)
			.. py:function:: typedlist(iterable, dtype, reserve=None)
			.. py:function:: typedlist(buffer, dtype)
			
		Attributes:
			
			size (int):        byte size of the current content
			allocated (int):   byte size of the memory allocated memory
			owner:             object realy owning the data instead of the current `typedlist`
			
			dtype (type):         the python data type
			ddtype (DDType):      the data type declaration
	'''

	cdef char *ptr
	cdef readonly size_t size
	cdef readonly size_t allocated
	
	cdef DDType dtype
	
	cdef readonly object owner
	
	@property
	def dtype(self):
		''' the python dtype object, or the ddtype if there is not dtype '''
		return self.dtype.key or self.dtype
		
	@property
	def ddtype(self):
		''' the declaration of the dtype, a DDType instance '''
		return self.dtype
	

	def __init__(self, iterable=None, object dtype=None, size_t reserve=0):
		# look at the type of the first element
		first = None
		if not dtype:
			try:
				if isinstance(iterable, type) and not dtype:
					iterable, dtype = None, iterable
				elif hasattr(iterable, '__next__'):
					first = next(iterable)
					dtype = type(first)
				elif hasattr(iterable, '__iter__'):
					dtype = type(next(iter(iterable)))
				else:
					raise TypeError('dtype must be provided when it cannot be deduced from the iterable')
			except StopIteration:
				raise ValueError('iterable is empty')
		
		# initialize the internal structure
		self.owner = None
		self.ptr = NULL
		self.size = 0
		self.allocated = 0
		self.dtype = declared(dtype)
		if not self.dtype:
			raise TypeError('dtype must be a dtype declaration instance, or a key for a declared dtype')
		
		# borrow a buffer
		if PyObject_CheckBuffer(iterable):
			self._use(iterable)
			self.size = self.allocated
			if reserve:
				self._reallocate(reserve * self.dtype.dsize)
		# fill with an iterable
		else:
			if reserve:
				self._reallocate(reserve * self.dtype.dsize)
			if iterable is not None:
				if first is not None:
					self.append(first)
				self.extend(iterable)
				
	@staticmethod
	def full(value, size, dtype=None):
		''' full(value, size)
		
			create a new typedlist with the given `size`, all elements initialized to `value`. 
		'''
		cdef ssize_t i
		cdef typedlist array = typedlist(dtype=dtype or type(value), reserve=size)
		for i in range(size):
			array._setitem(array.ptr + i*array.dtype.dsize, value)
		array.size = array.allocated
		return array
		
	@staticmethod
	def empty(dtype, size):
		''' empty(dtype, size)
		
			create a new typedlist with the given `size` and unitialized elements of type `dtype`
		'''
		cdef typedlist array = typedlist(dtype=dtype, reserve=size)
		array.size = array.allocated
		return array
	
	
	# convenient internal functions
	
	cdef object _getitem(self, void* place):
		return self.dtype.c_unpack(self.dtype, place)
		
	cdef int _setitem(self, void* place, obj) except -1:
		return self.dtype.c_pack(self.dtype, place, obj)
	
	cdef int _use(self, buffer) except -1:
		cdef Py_buffer view
		
		PyObject_GetBuffer(buffer, &view, PyBUF_SIMPLE)
		self.ptr = <char*> view.buf
		self.allocated = view.len
		self.owner = view.obj
		PyBuffer_Release(&view)
		
	cdef int _reallocate(self, size_t size) except -1:
		lastowner = self.owner
		lastptr = self.ptr
		
		# buffer protocol is less efficient that PyBytes_AS_STRING so we use it directly here where we know that owner is bytes
		self.owner = PyBytes_FromStringAndSize(NULL, size)
		self.ptr = PyBytes_AS_STRING(self.owner)
		#cdef buffer buff = buffer(size)
		#self.owner = buff
		#self.ptr = buff.ptr
		
		#print('** reallocate', sys.getrefcount(self.owner), sys.getrefcount(lastowner), lastowner is None)
		
		self.allocated = size
		self.size = min(size, self.size)
		memcpy(self.ptr, lastptr, self.size)
		
	cdef size_t _len(self):
		return self.size // self.dtype.dsize
		
	cdef Py_ssize_t _index(self, index) except -1:
		''' return a C index (0 < i < l) from a python object '''
		cdef Py_ssize_t i = index
		cdef Py_ssize_t l = self._len()
		if i < 0:	i += l
		if i < 0 or i >= l:	
			raise IndexError('index out of range')
		return i
	
	# python interface
	
	def reserve(self, amount):
		''' reserve(amount: int)
		
			Make sure there is enough allocated memory to append the given amount of elements. 
		
			if there is not enough of allocated memory, the memory is reallocated immediately.
		'''
		if amount < 0:	
			raise ValueError('amount must be positive')
		cdef size_t asked = self.size + (<size_t>amount) * self.dtype.dsize
		if asked > self.allocated:
			self._reallocate(asked)
	
	cpdef int append(self, value) except -1:
		''' append(value)
		
			append the given object at the end of the array
		
			if there is not enough allocated memory, reallocate enough to amortize the realocation time over the multiple appends
		'''		
		if self.allocated < self.size + self.dtype.dsize:
			self._reallocate(self.allocated*2 or self.dtype.dsize)
		
		self._setitem(self.ptr + self.size, value)
		self.size += self.dtype.dsize
		
	def pop(self, index=None):
		''' pop(index=None) -> object
		
			remove the element at index, returning it. If no index is specified, it will pop the last one .
		'''
		cdef size_t i
		cdef char * start
		
		if not self.size:
			raise IndexError('pop from empty list')
		
		if index is None:
			e = self._getitem(self.ptr + self.size - self.dtype.dsize)
			self.size -= self.dtype.dsize
			return e
		else:
			i = self._index(index)
			
			start = self.ptr + i*self.dtype.dsize
			e = self._getitem(start)
			memmove(start, start + self.dtype.dsize, self.size-(i+1)*self.dtype.dsize)
			self.size -= self.dtype.dsize
			return e
		
	def insert(self, index, value):
		''' insert(index, value)
		
			insert value at index 
		'''
		
		cdef Py_ssize_t i = index
		cdef Py_ssize_t l = self._len()
		if i < 0:	i += l
		if i < 0 or i > l:	
			raise IndexError('index out of range')
		cdef size_t j = i
		
		if self.allocated - self.size < self.dtype.dsize:
			self._reallocate(self.allocated*2 or self.dtype.dsize)
			
		cdef char * start = self.ptr + j*self.dtype.dsize
		memmove(start+self.dtype.dsize, start, self.size-j*self.dtype.dsize)
		self._setitem(start, value)
		self.size += self.dtype.dsize
		
	def clear(self):
		''' remove all elements from the array but does not deallocate, very fast operation '''
		self.size = 0
		
	cpdef int extend(self, other) except *:
		''' extend(iterable)
		
			append all elements from the other array 
		'''
		cdef Py_buffer view
		cdef Py_ssize_t l
		
		if PyObject_CheckBuffer(other):
			PyObject_GetBuffer(other, &view, PyBUF_SIMPLE)
			
			if view.len % self.dtype.dsize:
				PyBuffer_Release(&view)
				raise TypeError('the given buffer must have a byte size multiple of dtype size')
			
			if <size_t>view.len > self.allocated - self.size:
				self._reallocate(max(2*self.size, self.size + view.len))
			memcpy(self.ptr+self.size, view.buf, view.len)
			self.size += view.len
			
			PyBuffer_Release(&view)
			
		else:
			l = PyObject_LengthHint(other, 0)
			if l*self.dtype.dsize > self.allocated - self.size:
				self._reallocate(self.size + l*self.dtype.dsize)
			for o in other:
				self.append(o)
				
	def __iadd__(self, other):
		self.extend(other)
		return self
		
	def __add__(typedlist self, other):
		''' concatenation of two arrays '''
		cdef typedlist result
		cdef Py_buffer view
		
		if PyObject_CheckBuffer(other):
			PyObject_GetBuffer(other, &view, PyBUF_SIMPLE)
			
			if view.len % self.dtype.dsize:
				PyBuffer_Release(&view)
				raise TypeError('the given buffer must have a byte size multiple of dtype size')
			
			result = typedlist.__new__(typedlist)
			result.dtype = self.dtype
			result._reallocate(self.size + view.len)
			memcpy(result.ptr, self.ptr, self.size)
			memcpy(result.ptr+self.size, view.buf, view.len)
			result.size = result.allocated
			
			PyBuffer_Release(&view)
			return result
			
		elif hasattr(other, '__iter__'):
			result = typedlist(None, self.dtype)
			result.extend(self)
			result.extend(other)
			return result
			
		else:
			return NotImplemented
			
	def __mul__(self, n):
		''' duplicate the sequence by a certain number '''
		#if isinstance(n, int):
		if PyNumber_Check(n):
			return typedlist(bytes(self)*n, dtype=self.dtype)
		else:
			return NotImplemented
	
	def capacity(self):
		''' capacity()
		
			return the total number of elements that can be stored with the current allocated memory 
		'''
		return self.allocated // self.dtype.dsize
		
	def shrink(self):
		''' shrink()
		
			reallocate the array to have allocated the exact current size of the array 
		'''
		self._reallocate(self.size)
		
	def __len__(self):
		''' return the current amount of elements inserted '''
		return self._len()
		
	def __getitem__(self, index):
		''' self[index]
		
			currently supports:
			
				- indices
				- negative indices
				- slices with step=1
		'''
		cdef typedlist view
		cdef Py_ssize_t start, stop, step
		
		#if isinstance(index, int):
		if PyNumber_Check(index):
			return self._getitem(self.ptr + self._index(index)*self.dtype.dsize)
		
		elif isinstance(index, slice):
			if PySlice_Unpack(index, &start, &stop, &step):
				raise IndexError('incorrect slice')
			if step != 1:
				raise IndexError('slice step is not supported')
			PySlice_AdjustIndices(self._len(), &start, &stop, step)
			if stop < start:
				stop = start
			
			view = typedlist.__new__(typedlist)
			view.ptr = self.ptr + start*self.dtype.dsize
			view.size = (stop-start)*self.dtype.dsize
			view.allocated = view.size
			view.owner = self.owner
			view.dtype = self.dtype
			return view
		
		else:
			raise IndexError('index must be int or slice')
			
	def __setitem__(self, index, value):
		cdef Py_buffer view
		cdef Py_ssize_t start, stop, step
		cdef size_t newsize
		
		#if isinstance(index, int):
		if PyNumber_Check(index):
			self._setitem(self.ptr + self._index(index)*self.dtype.dsize, value)
			
		elif isinstance(index, slice):
			if PySlice_Unpack(index, &start, &stop, &step):
				raise IndexError('incorrect slice')
			if step != 1:
				raise IndexError('slice step is not supported')
			PySlice_AdjustIndices(self._len(), &start, &stop, step)
			if stop < start:
				stop = start
			
			if PyObject_CheckBuffer(value):
				PyObject_GetBuffer(value, &view, PyBUF_SIMPLE)
				start *= self.dtype.dsize
				stop *= self.dtype.dsize
				
				if view.len != stop-start:
					if view.len % self.dtype.dsize:
						PyBuffer_Release(&view)
						raise TypeError('the given buffer must have a byte size multiple of dtype size')
					
					newsize = self.size + view.len + start - stop
					if newsize >  self.allocated:
						self._reallocate(max(2*self.size, newsize))
					
					memmove(self.ptr+start+view.len, self.ptr+stop, self.size-stop)
					self.size = newsize
				
				memcpy(self.ptr+start, view.buf, view.len)
				
				PyBuffer_Release(&view)
				
			elif hasattr(value, '__iter__'):
				self[index] = typedlist(value, self.dtype)
				
			else:
				raise IndexError('the assigned value must be a buffer or an iterable')
			
		else:
			raise IndexError('index must be int or slice')
			
	def __delitem__(self, index):
		cdef Py_ssize_t start, stop, step
		
		if PyNumber_Check(index):
			start = index
			stop = index+1
		elif isinstance(index, slice):
			if PySlice_Unpack(index, &start, &stop, &step):
				raise IndexError('incorrect slice')
			if step != 1:
				raise IndexError('slice step is not supported')
			PySlice_AdjustIndices(self._len(), &start, &stop, step)
		else:
			raise IndexError('index must be int or slice')
			
		start *= self.dtype.dsize
		stop *= self.dtype.dsize
		memmove(self.ptr+start, self.ptr+stop, self.size-stop)
		self.size = self.size + start - stop

			
	def __iter__(self):
		''' yield successive elements in the list '''
		cdef arrayiter it = arrayiter.__new__(arrayiter)
		it.array = self
		it.position = 0
		return it
		
	def __repr__(self):
		cdef size_t i
		text = 'typedlist(['
		for i in range(self._len()):
			if i:	text += ', '
			text += repr(self._getitem(self.ptr + i*self.dtype.dsize))
		text += '])'
		return text
		
	def __eq__(self, other):
		''' return True if other is a typedlist and its buffer byte contents is the same '''
		if not isinstance(other, typedlist):	return False
		cdef typedlist o = other
		if self.size != o.size:		return False
		return 0 == memcmp(self.ptr, o.ptr, self.size)
		
	def __copy__(self):
		''' shallow copy will create a copy of that array referencing the same buffer '''
		return self[:]
		
	def __deepcopy__(self, memo):
		''' deep recursive copy,  will duplicate the viewed data in the underlying buffer '''
		return typedlist(PyBytes_FromStringAndSize(<char*>self.ptr, self.size), self.dtype)
		
	def __reduce_ex__(self, protocol):
		''' serialization protocol '''
		cdef Py_buffer view
		
		if protocol >= 5:
			if not PyObject_CheckBuffer(self.owner):
				raise RuntimeError("the buffer owner doens't implement the buffer protocol")
			PyObject_GetBuffer(self.owner, &view, PyBUF_SIMPLE)
			stuff = (
						PickleBuffer(self.owner), 
						self.dtype, 
						self.ptr - <char*>view.buf, 
						self.size,
						)
			PyBuffer_Release(&view)
			return self._rebuild, stuff, None
		else:
			return self._rebuild, (
						PyBytes_FromStringAndSize(<char*>self.ptr, self.size), 
						self.dtype,
						0, 
						self.size,
						)
	
	@classmethod
	def _rebuild(cls, owner, dtype, size_t start, size_t size):
		new = typedlist(owner, dtype)
		assert start <= size
		assert size <= new.size
		new.ptr = new.ptr + start
		new.size = size
		return new
		
	def __getbuffer__(self, Py_buffer *view, int flags):
		cdef arrayexposer exp = arrayexposer.__new__(arrayexposer)
		exp.owner = self.owner
		exp.shape[0] = self._len()
	
		assign_buffer_obj(view, None)
		view.obj = exp
		view.buf = self.ptr
		view.len = self.size
		view.ndim = 1
		
		if flags & PyBUF_FORMAT:
			fmt = self.dtype.layout
			if fmt is not None:
				view.itemsize = self.dtype.dsize
				view.format = PyBytes_AS_STRING(fmt)
			else:
				view.itemsize = 1
				view.format = 'B'
		else:
			view.itemsize = 1
			view.format = NULL
		
		if flags & PyBUF_ND:
			view.suboffsets = NULL
			view.strides = &view.itemsize
			view.shape = exp.shape
		else:
			view.shape = NULL
			
		
	def __releasebuffer__(self, Py_buffer *view):
		cdef arrayexposer exp = view.obj
		print('** release buffer', sys.getrefcount(exp.owner))
		exp.owner = None
		view.obj = None
		
	def reverse(self):
		''' reverse()
		
			reverse the order of elementd contained 
		'''
		cdef char *temp
		cdef char *first
		cdef char *last
		cdef size_t i
		
		temp = <char*> PyMem_Malloc(self.dtype.dsize)
		first = self.ptr
		last = self.ptr + self.size - self.dtype.dsize
		while <size_t>first < <size_t>last:
			memcpy(temp, first, self.dtype.dsize)
			memcpy(first, last, self.dtype.dsize)
			memcpy(last, temp, self.dtype.dsize)
			first += self.dtype.dsize
			last -= self.dtype.dsize
		PyMem_Free(temp)
		
	def index(self, value):
		''' index(value)
		
			return the index of the first element binarily equal to the given one 
		'''
		cdef size_t i, j
		cdef char *data
		cdef char *val
			
		data = <char*> self.ptr
		val = <char*> PyMem_Malloc(self.dtype.dsize)
		self._setitem(val, value)
		
		i = 0
		while i < self.size:
			j = 0
			while data[i+j] == val[j] and j < self.dtype.dsize:
				j += 1
			if j == self.dtype.dsize:
				PyMem_Free(val)
				return i//self.dtype.dsize
			i += self.dtype.dsize
			
		PyMem_Free(val)
		raise IndexError('value not found')
			
			
			
cdef class arrayexposer:
	''' very simple object that just holds the data for the buffer objects '''
	cdef readonly object owner
	cdef Py_ssize_t shape[1]



cdef class arrayiter:
	cdef typedlist array
	cdef size_t position
	
	def __init__(self):
		raise TypeError('arrayiter must not be instanciated explicitely')
	
	def __iter__(self):
		return self
		
	def __next__(self):
		if self.position + self.array.dtype.dsize > self.array.size:
			raise StopIteration
		item = self.array._getitem(self.array.ptr + self.position)
		self.position += self.array.dtype.dsize
		return item

		
		

'''	
# this is in reserve for debug purpose
# helps to keep track of buffers

cdef size_t buffer_id = 0
cdef size_t buffer_count = 0

cdef class buffer:
	cdef void* ptr
	cdef readonly size_t size
	cdef size_t id
	
	def __cinit__(self, size_t size):
		global buffer_id, buffer_count
		self.id = buffer_id
		buffer_id += 1
		buffer_count += 1
		print('allocate', self.id, size)
		
		self.ptr = PyMem_Malloc(size)
		self.size = size
		
	def __dealloc__(self):
		global buffer_id, buffer_count
		buffer_count -= 1
		PyMem_Free(self.ptr)
		self.ptr = NULL
		print('deallocate', self.id, buffer_count)
		
	def __len__(self):
		return self.size
'''



# cython: language_level=3, cdivision=True

cimport cython
from cpython cimport PyObject, Py_DECREF
from libc.stdint cimport *
from .dtypes cimport *

cdef DDType decl


### declare double

cdef int pack_d(object dtype, double* place, object obj) except -1:
	place[0] = obj
cdef object unpack_d(object dtype, double* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(double)
decl.c_pack = <c_pack_t> pack_d
decl.c_unpack = <c_unpack_t> unpack_d
decl.layout = b'd'

declare('d', decl)


### declare float

cdef int pack_f(object dtype, float* place, object obj) except -1:
	place[0] = obj
cdef object unpack_f(object dtype, float* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(float)
decl.c_pack = <c_pack_t> pack_f
decl.c_unpack = <c_unpack_t> unpack_f
decl.layout = b'f'

declare('f', decl)


### declare int8_t

cdef int pack_b(object dtype, int8_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_b(object dtype, int8_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(int8_t)
decl.c_pack = <c_pack_t> pack_b
decl.c_unpack = <c_unpack_t> unpack_b
decl.layout = b'b'

declare('b', decl)


### declare uint8_t

cdef int pack_B(object dtype, uint8_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_B(object dtype, uint8_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(uint8_t)
decl.c_pack = <c_pack_t> pack_B
decl.c_unpack = <c_unpack_t> unpack_B
decl.layout = b'B'

declare('B', decl)


### declare int16_t

cdef int pack_h(object dtype, int16_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_h(object dtype, int16_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(int16_t)
decl.c_pack = <c_pack_t> pack_h
decl.c_unpack = <c_unpack_t> unpack_h
decl.layout = b'h'

declare('h', decl)


### declare uint16_t

cdef int pack_H(object dtype, uint16_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_H(object dtype, uint16_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(uint16_t)
decl.c_pack = <c_pack_t> pack_H
decl.c_unpack = <c_unpack_t> unpack_H
decl.layout = b'H'

declare('H', decl)


### declare int32_t

cdef int pack_i(object dtype, int32_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_i(object dtype, int32_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(int32_t)
decl.c_pack = <c_pack_t> pack_i
decl.c_unpack = <c_unpack_t> unpack_i
decl.layout = b'i'

declare('i', decl)


### declare uint32_t

cdef int pack_I(object dtype, uint32_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_I(object dtype, uint32_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(uint32_t)
decl.c_pack = <c_pack_t> pack_I
decl.c_unpack = <c_unpack_t> unpack_I
decl.layout = b'I'

declare('I', decl)


### declare int64_t

cdef int pack_l(object dtype, int64_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_l(object dtype, int64_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(int64_t)
decl.c_pack = <c_pack_t> pack_l
decl.c_unpack = <c_unpack_t> unpack_l
decl.layout = b'l'

declare('l', decl)


### declare uint64_t

cdef int pack_L(object dtype, uint64_t* place, object obj) except -1:
	place[0] = obj
cdef object unpack_L(object dtype, uint64_t* place):
	return place[0]

decl = DDType()
decl.dsize = sizeof(uint64_t)
decl.c_pack = <c_pack_t> pack_L
decl.c_unpack = <c_unpack_t> unpack_L
decl.layout = b'L'

declare('L', decl)

declare(float, declared('d'))
declare(int, declared('l'))



import numpy.core as np
from .dtypes import declare, DDTypeExtension

f16 = np.float16
f32 = np.float32
f64 = np.float64
f128 = np.float128

u8 = np.uint8
u16 = np.uint16
u32 = np.uint32
u64 = np.uint64

i8 = np.int8
i16 = np.int16
i32 = np.int32
i64 = np.int64

declare(f16, DDTypeExtension(f16, 'exxxxxx', f16))
declare(f32, DDTypeExtension(f32, 'fxxxx', f32))
declare(f64, DDTypeExtension(f64, 'd', f64))
declare(f128, DDTypeExtension(f128, 'B'*16, f128))

declare(u8, DDTypeExtension(u8, 'Bxxxxxxx', u8))
declare(i8, DDTypeExtension(i8, 'bxxxxxxx', i8))
declare(u16, DDTypeExtension(u16, 'Hxxxxxx', u16))
declare(i16, DDTypeExtension(i16, 'hxxxxxx', i16))
declare(u32, DDTypeExtension(u32, 'Ixxxx', u32))
declare(i16, DDTypeExtension(i32, 'ixxxx', i32))
declare(u64, DDTypeExtension(u64, 'L', u64))
declare(i64, DDTypeExtension(i64, 'l', i64))



''' Arrex is a module that allows to create typed arrays much like numpy.ndarray and array.array, but resizeable and using any kind of element.

	The elements must be extension-types (eg. class created in compiled modules) and must have a packed and copyable content: a fixed size and no reference or pointers
	This is meant to ensure that the content of those objects can be copied from the object to the array and back to any object after, or even deleted without any need of calling a constructor or destructor function.
	
	basic usage:
	
		>>> from arrex import *
		>>> a = typedlist([
		...             myclass(...), 
		...             myclass(...),
		...             ], dtype=myclass)
		>>> a[0]
		myclass(...)
	
	in that example, `myclass` can be a primitive numpy type, like `np.float64`
	
		>>> import typedlist.numpy		# this is enabling numpy dtypes for arrex
		>>> typedlist(dtype=np.float64)
		
	it can be a more complex type, from module `pyglm` for instance
	
		>>> import typedlist.glm		# this is enabling glm dtypes for arrex
		>>> typedlist(dtype=glm.vec4)
		
		>>> a = typedlist(dtype=vec3)

	use it as a list
	
		>>> # build from an iterable
		>>> a = typedlist([], dtype=vec3)
		>>>
		>>> # append some data
		>>> a.append(vec3(1,2,3))
		>>>
		>>> # extend with an iterable
		>>> a.extend(vec3(i)  for i in range(5))
		>>>
		>>> len(a)	# the current number of elements
		6
		>>> a.owner	# the current data buffer
		b'.........'
		>>> a[0]
		vec3(1,2,3)
		
		
	Use it as a slice:

		>>> myslice = a[:5]		# no data is copied
		typedlist(....)
	
	Use it as a view on top of a random buffer

		>>> a = np.ones((6,3), dtype='f4')
		>>> myslice = typedlist(a, dtype=vec3)
	
	It does support the buffer protocol, so it can be converted into a great variety of well known arrays, even without any copy

		>>> np.array(typedlist([....]))
	
'''

__all__ = ['typedlist']

from .dtypes import *
from .list import typedlist
from . import numbers



# cython: language_level=3, cdivision=True

''' 
	This is the header file for the arrex cython-only API. 
	Of course you can use the python API to complete what is here.
'''

cimport cython
from cpython cimport PyObject

ctypedef int (*c_pack_t) (object, void*, object) except -1
ctypedef object (*c_unpack_t) (object, void*)

cdef class DDType:
	''' base class for a dtype, But you should use on of its specialization instead '''
	cdef public size_t dsize	# size of one element in a buffer
	cdef c_pack_t c_pack  		# pointer to the function that packs the python object into a buffer
	cdef c_unpack_t c_unpack 	# pointer to the function that create the python object using a buffer content
	cdef public bytes layout    # (optional)  element layout following specifications in module `struct`
	cdef public object key      # value to return when one asks an array for its dtype (it can be the current class itself, or any hashable value representing it)
	
	'''
	Just a piece of recommendation:
		
	- the method/function used to put in c_pack and c_unpack, MUST use type <object> as argument and return type, or it will wrongly increment/decrement its refcount and make a memory leak or a segfault
	'''

	
cpdef DDType declare(key, DDType dtype=*)
cpdef DDType declared(key)



from setuptools import setup, Extension
import os

try:
	from Cython.Build import cythonize
except ImportError:
	cython_modules = [
				Extension('arrex.dtypes', ['arrex/dtypes.c']),
				Extension('arrex.list', ['arrex/list.c']),
				Extension('arrex.numbers', ['arrex/numbers.c']),
				]
else:
	cython_modules = cythonize(['arrex/dtypes.pyx', 'arrex/list.pyx', 'arrex/numbers.pyx'], annotate=True)

setup(
	# package declaration
	name = 'arrex',
	version = '0.5.0',
	python_requires='>=3.8',
	tests_require = [
		'pnprint>=1.1',
		'pyglm>=1.2',
		'numpy>=1.1',
		],
	
	# sources declaration
	packages = ["arrex"],
	package_data = {
		"arrex": ['*.h', '*.c', '*.cpp', '*.pyx', '*.pxd'],
		'': ['COPYING', 'COPYING.LESSER', 'README'],
		},
	ext_modules = cython_modules,
	
	# metadata for pypi
	author = 'Jimy Byerley',
	author_email = 'jimy.byerley@gmail.com',
	url = 'https://github.com/jimy-byerley/arrex',
	license = "GNU LGPL v3",
	description = "typed arrays using any custom type as element type",
	long_description = open('README.md').read(),
	long_description_content_type = 'text/markdown',
	keywords='buffer array list dynamic dtype serialization numeric',
	classifiers=[
		'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
		'Development Status :: 3 - Alpha',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Programming Language :: Python :: 3.10',
		'Topic :: Software Development :: Libraries',
		'Topic :: Utilities',
		],
)
